"""Module to provide functionality to import structures."""

import datetime
import functools
import io
import pathlib
import tempfile

import ase
import ipywidgets as ipw
import numpy as np
import spglib
import traitlets as tl
from aiida import engine, orm, plugins

# Local imports
from .data import FunctionalGroupSelectorWidget
from .utils import StatusHTML, exceptions, get_ase_from_file, get_formula
from .viewers import StructureDataViewer

CifData = plugins.DataFactory("core.cif")
StructureData = plugins.DataFactory("core.structure")
TrajectoryData = plugins.DataFactory("core.array.trajectory")

SYMBOL_RADIUS = {
    key: ase.data.covalent_radii[i] for i, key in enumerate(ase.data.chemical_symbols)
}


class StructureManagerWidget(ipw.VBox):
    """Upload a structure and store it in AiiDA database.

    Attributes:
        structure(Atoms): trait that contains the selected structure. 'None' if no structure is selected.
        structure_node(StructureData, CifData): trait that contains AiiDA structure object
        node_class(str): trait that contains structure_node type (as string).
    """

    input_structure = tl.Union(
        [tl.Instance(ase.Atoms), tl.Instance(orm.Data)], allow_none=True
    )
    structure = tl.Instance(ase.Atoms, allow_none=True)
    structure_node = tl.Instance(orm.Data, allow_none=True, read_only=True)
    node_class = tl.Unicode()

    SUPPORTED_DATA_FORMATS = {"CifData": "core.cif", "StructureData": "core.structure"}

    def __init__(
        self,
        importers,
        viewer=None,
        editors=None,
        storable=True,
        node_class=None,
        **kwargs,
    ):
        """
        Arguments:
            importers(list): list of tuples each containing the displayed name of importer and the
                importer object. Each object should contain 'structure' trait pointing to the imported
                structure. The trait will be linked to 'structure' trait of this class.

            storable(bool): Whether to provide Store button (together with Store format)

            node_class(str): AiiDA node class for storing the structure.
                Possible values: 'StructureData', 'CifData' or None (let the user decide).
                Note: If your workflows require a specific node class, better fix it here.
        """

        # History of modifications
        self.history = []

        # Undo functionality.
        btn_undo = ipw.Button(description="Undo", button_style="success")
        btn_undo.on_click(self.undo)
        self.structure_set_by_undo = False

        # To keep track of last inserted structure object
        self._inserted_structure = None

        # Structure viewer.
        if viewer:
            self.viewer = viewer
        else:
            self.viewer = StructureDataViewer()
        tl.dlink((self, "structure"), (self.viewer, "structure"))

        # Store button.
        self.btn_store = ipw.Button(description="Store in AiiDA", disabled=True)
        self.btn_store.on_click(self.store_structure)

        # Label and description that are stored along with the new structure.
        self.structure_label = ipw.Text(description="Label")
        self.structure_description = ipw.Text(description="Description")

        # Store format selector.
        data_format = ipw.RadioButtons(
            options=tuple(
                (key, value) for key, value in self.SUPPORTED_DATA_FORMATS.items()
            ),
            description="Data type:",
        )
        tl.link((data_format, "label"), (self, "node_class"))

        # Store button, store class selector, description.
        store_and_description = [self.btn_store] if storable else []

        if node_class is None:
            store_and_description.append(data_format)
        elif node_class in self.SUPPORTED_DATA_FORMATS:
            self.node_class = node_class
        else:
            raise ValueError(
                f"Unknown data format '{node_class}'. Options: {list(self.SUPPORTED_DATA_FORMATS.keys())}"
            )

        select_panel = ipw.Accordion(
            children=self._structure_importers(importers),
            selected_index=0 if importers else None,
        )
        select_panel.set_title(0, "Select structure")

        view_panel = ipw.Accordion(
            children=[
                ipw.VBox(
                    children=[
                        self.viewer,
                        ipw.HBox(
                            children=[
                                *store_and_description,
                                self.structure_label,
                                self.structure_description,
                            ]
                        ),
                    ]
                ),
            ],
            selected_index=0,
        )
        view_panel.set_title(0, "View structure")

        structure_editors = self._structure_editors(editors)

        edit_panel = ipw.Accordion(
            children=[
                ipw.VBox(
                    children=[
                        btn_undo,
                        *structure_editors,
                    ]
                ),
            ]
            if structure_editors
            else [],
            selected_index=None,
        )
        edit_panel.set_title(0, "Edit structure")

        self.output = ipw.HTML("")

        super().__init__(
            children=[
                select_panel,
                view_panel,
                edit_panel,
                self.output,
            ],
            **kwargs,
        )

        self.add_class("structure-manager")

    def _structure_importers(self, importers):
        """Preparing structure importers."""
        if not isinstance(importers, (list, tuple)):
            raise exceptions.ListOrTuppleError(importers)

        if not importers:
            return []

        # Otherwise making one tab per importer.
        importers_tab = ipw.Tab()
        importers_tab.children = list(importers)  # One importer per tab.
        for i, importer in enumerate(importers):
            # Labeling tabs.
            importers_tab.set_title(i, importer.title)
            tl.dlink((importer, "structure"), (self, "input_structure"))
        return [importers_tab]

    def _structure_editors(self, editors):
        """Preparing structure editors."""
        if not editors:
            return []

        editors_tab = ipw.Tab()
        editors_tab.children = tuple(editors)
        for i, editor in enumerate(editors):
            editors_tab.set_title(i, editor.title)
            # Link selected traits of the editors with the those of the viewer.
            tl.link((editor, "structure"), (self, "structure"))
            if editor.has_trait("input_selection"):
                tl.dlink((editor, "input_selection"), (self.viewer, "input_selection"))
            if editor.has_trait("selection"):
                tl.link((editor, "selection"), (self.viewer, "selection"))
            if editor.has_trait("camera_orientation"):
                tl.dlink(
                    (self.viewer._viewer, "_camera_orientation"),
                    (editor, "camera_orientation"),
                )

        return [editors_tab]

    def store_structure(self, _=None):
        """Stores the structure in AiiDA database."""

        if self.structure_node is None:
            return
        if self.structure_node.is_stored:
            self.output.value = (
                f"Already stored in AiiDA [{self.structure_node}], skipping..."
            )
            return
        self.btn_store.disabled = True
        self.structure_node.label = self.structure_label.value
        self.structure_label.disabled = True
        self.structure_node.description = self.structure_description.value
        self.structure_description.disabled = True

        if (
            isinstance(self.input_structure, orm.Data)
            and self.input_structure.is_stored
        ):
            # Make a link between self.input_structure and self.structure_node
            @engine.calcfunction
            def user_modifications(source_structure):  # noqa F841
                return self.structure_node

            structure_node = user_modifications(self.input_structure)

        else:
            structure_node = self.structure_node.store()
        self.output.value = f"Stored in AiiDA [{structure_node}]"

    def undo(self, _=None):
        """Undo modifications."""
        self.structure_set_by_undo = True
        if self.history:
            self.history = self.history[:-1]
            if self.history:
                self.structure = self.history[-1]
            else:
                self.input_structure = None
        self.structure_set_by_undo = False

    @staticmethod
    @tl.default("node_class")
    def _default_node_class():
        return "StructureData"

    @tl.observe("node_class")
    def _change_structure_node(self, _=None):
        with self.hold_trait_notifications():
            self._sync_structure_node()

    def _sync_structure_node(self):
        """Synchronize the structure_node trait using the currently provided info."""
        if len(self.history) > 1:
            # There are some modifications, so converting from ASE.
            structure_node = self._convert_to_structure_node(self.structure)
        else:
            structure_node = self._convert_to_structure_node(self.input_structure)
        self.set_trait("structure_node", structure_node)

    def _convert_to_structure_node(self, structure):
        """Convert structure of any type to the StructureNode object."""
        if structure is None:
            return None
        structure_node_type = plugins.DataFactory(
            self.SUPPORTED_DATA_FORMATS[self.node_class]
        )

        # If the input_structure trait is set to Atoms object, structure node must be created from it.
        if isinstance(structure, ase.Atoms):
            # If the Atoms object was created by SmilesWidget,
            # attach its SMILES code as an extra.
            structure_node = structure_node_type(ase=structure)
            if "smiles" in structure.info:
                structure_node.base.extras.set("smiles", structure.info["smiles"])
            return structure_node

        # If the input_structure trait is set to AiiDA node, check what type
        if isinstance(structure, orm.Data):
            # Transform the structure to the structure_node_type if needed.
            if isinstance(structure, structure_node_type):
                return structure

        # Using self.structure, as it was already converted to the ASE Atoms object.
        return structure_node_type(ase=self.structure)

    @tl.observe("structure_node")
    def _observe_structure_node(self, change):
        """Modify structure label and description when a new structure is provided."""
        struct = change["new"]
        if struct is None:
            self.btn_store.disabled = True
            self.structure_label.value = ""
            self.structure_label.disabled = True
            self.structure_description.value = ""
            self.structure_description.disabled = True
            return
        if struct.is_stored:
            self.btn_store.disabled = True
            self.structure_label.value = struct.label
            self.structure_label.disabled = True
            self.structure_description.value = struct.description
            self.structure_description.disabled = True
        else:
            self.btn_store.disabled = False
            self.structure_label.value = self.structure.get_chemical_formula()
            self.structure_label.disabled = False
            self.structure_description.value = ""
            self.structure_description.disabled = False

    @tl.observe("input_structure")
    def _observe_input_structure(self, change):
        """Returns ASE atoms object and sets structure_node trait."""
        # If the `input_structure` trait is set to Atoms object, then the `structure` trait should be set to it as well.
        self.history = []

        if isinstance(change["new"], ase.Atoms):
            self.structure = change["new"]

        # If the `input_structure` trait is set to AiiDA node, then the `structure` trait should
        # be converted to an ASE Atoms object.
        elif isinstance(
            change["new"], CifData
        ):  # Special treatement of the CifData object
            str_io = io.StringIO(change["new"].get_content())
            self.structure = ase.io.read(
                str_io, format="cif", reader="ase", store_tags=True
            )
        elif isinstance(change["new"], StructureData):
            self.structure = change["new"].get_ase()

        else:
            self.structure = None

    @tl.observe("structure")
    def _structure_changed(self, change=None):
        """Perform some operations that depend on the value of `structure` trait.

        This function enables/disables `btn_store` widget if structure is provided/set to None.
        Also, the function sets `structure_node` trait to the selected node type.
        """

        if not self.structure_set_by_undo:
            self.history.append(change["new"])

        # If structure trait was set to None, structure_node should become None as well.
        if self.structure is None:
            self.set_trait("structure_node", None)
            self.btn_store.disabled = True
            return

        self.btn_store.disabled = False
        with self.hold_trait_notifications():
            self._sync_structure_node()


class StructureUploadWidget(ipw.VBox):
    """Class that allows to upload structures from user's computer."""

    structure = tl.Union(
        [tl.Instance(ase.Atoms), tl.Instance(orm.Data)], allow_none=True
    )

    def __init__(
        self, title="", description="Upload Structure", allow_trajectories=False
    ):
        self.title = title
        self.file_upload = ipw.FileUpload(
            description=description, multiple=False, layout={"width": "initial"}
        )
        # Whether to allow uploading multiple structures from a single file.
        # In this case, we create TrajectoryData node.
        self.allow_trajectories = allow_trajectories
        supported_formats = ipw.HTML(
            """<a href="https://wiki.fysik.dtu.dk/ase/ase/io/io.html#ase.io.write" target="_blank">
        Supported structure formats
        </a>"""
        )
        self._status_message = StatusHTML(clear_after=5)
        self.file_upload.observe(self._on_file_upload, names="value")
        super().__init__(
            children=[self.file_upload, supported_formats, self._status_message]
        )

    def _validate_and_fix_ase_cell(self, ase_structure, vacuum_ang=10.0):
        """
        Checks if the ase Atoms object has a cell set,
        otherwise sets it to bounding box plus specified "vacuum" space
        """
        if not ase_structure:
            return None

        cell = ase_structure.cell

        # TODO: Since AiiDA 2.0, zero cell is possible if PBC=false
        # so we should honor that here and do not put artificial cell
        # around gas phase molecules.
        if (
            np.linalg.norm(cell[0]) < 0.1
            or np.linalg.norm(cell[1]) < 0.1
            or np.linalg.norm(cell[2]) < 0.1
        ):
            # if any of the cell vectors is too short, consider it faulty
            # set cell as bounding box + vacuum_ang
            bbox = np.ptp(ase_structure.positions, axis=0)
            new_structure = ase_structure.copy()
            new_structure.cell = bbox + vacuum_ang
            return new_structure
        return ase_structure

    def _on_file_upload(self, change=None):
        """When file upload button is pressed."""
        for fname, item in change["new"].items():
            self.structure = self._read_structure(fname, item["content"])
            self.file_upload.value.clear()
            break

    def _read_structure(self, fname, content):
        suffix = "".join(pathlib.Path(fname).suffixes)
        if suffix == ".cif":
            try:
                return CifData(file=io.BytesIO(content))
            except Exception as e:
                self._status_message.message = f"""
                    <div class="alert alert-warning">Could not parse CIF file {fname}: {e}
                    Trying ASE reader...</div>
                    """

        with tempfile.NamedTemporaryFile(suffix=suffix) as temp_file:
            temp_file.write(content)
            temp_file.flush()
            try:
                if suffix == ".cif":
                    structures = get_ase_from_file(temp_file.name, format="cif")
                else:
                    structures = get_ase_from_file(temp_file.name)
            except ValueError as e:
                self._status_message.message = f"""
                    <div class="alert alert-danger">ERROR: {e}</div>
                    """
                return None
            except KeyError:
                self._status_message.message = f"""
                    <div class="alert alert-danger">ERROR: Could not parse file {fname}</div>
                    """
                return None

            if len(structures) > 1:
                if self.allow_trajectories:
                    return TrajectoryData(
                        structurelist=[
                            StructureData(
                                ase=self._validate_and_fix_ase_cell(ase_struct)
                            )
                            for ase_struct in structures
                        ]
                    )
                else:
                    self._status_message.message = f"""
                        <div class="alert alert-danger">ERROR: More than one structure found in file {fname}</div>
                        """
                    return None

            return self._validate_and_fix_ase_cell(structures[0])


class StructureExamplesWidget(ipw.VBox):
    """Class to provide example structures for selection."""

    structure = tl.Instance(ase.Atoms, allow_none=True)

    def __init__(self, examples, title="", **kwargs):
        self.title = title
        self.on_structure_selection = lambda _structure_ase, _name: None
        self._select_structure = ipw.Dropdown(
            options=self.get_example_structures(examples)
        )
        self._select_structure.observe(self._on_select_structure, names=["value"])
        super().__init__(children=[self._select_structure], **kwargs)

    @staticmethod
    def get_example_structures(examples):
        """Get the list of example structures."""
        if not isinstance(examples, list):
            raise TypeError(
                f"parameter examples should be of type list, {type(examples)} given"
            )
        return [("Select structure", False), *examples]

    def _on_select_structure(self, change=None):
        """When structure is selected."""

        self.structure = (
            get_ase_from_file(self._select_structure.value)[0]
            if self._select_structure.value
            else None
        )

    @tl.default("structure")
    def _default_structure(self):
        return None


class StructureBrowserWidget(ipw.VBox):
    """Class to query for structures stored in the AiiDA database.

    :param title: Title of the widget displayed on a tab in StructureManagerWidget
    :type title: string
    :param query_types: A tuple of Data node types that are searched (default: StructureData, CifData)
    :type query_types: tuple
    """

    structure = tl.Union(
        [tl.Instance(ase.Atoms), tl.Instance(orm.Data)], allow_none=True
    )

    def __init__(self, title="", query_types=None):
        self.title = title

        # Structure objects we want to query for.
        if query_types:
            self.query_structure_type = query_types
        else:
            self.query_structure_type = (StructureData, CifData)

        # Extracting available process labels.
        qbuilder = orm.QueryBuilder().append(
            (orm.CalcJobNode, orm.WorkChainNode), project="label"
        )
        self.drop_label = ipw.Dropdown(
            options=sorted({"All"}.union({i[0] for i in qbuilder.iterall() if i[0]})),
            value="All",
            description="Process Label",
            disabled=True,
            style={"description_width": "120px"},
            layout={"width": "50%"},
        )
        self.drop_label.observe(self.search, names="value")

        # Disable process labels selection if we are not looking for the calculated structures.
        def disable_drop_label(change):
            self.drop_label.disabled = not change["new"] == "calculated"

        # Select structures kind.
        self.mode = ipw.RadioButtons(
            options=["all", "uploaded", "edited", "calculated"], layout={"width": "25%"}
        )
        self.mode.observe(self.search, names="value")
        self.mode.observe(disable_drop_label, names="value")

        # Date range.
        # Note: there is Date picker widget, but it currently does not work in Safari:
        # https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20List.html#Date-picker
        date_text = ipw.HTML(value="<p>Select the date range:</p>")
        self.start_date_widget = ipw.Text(
            value="", description="From: ", style={"description_width": "120px"}
        )
        self.end_date_widget = ipw.Text(value="", description="To: ")

        # Search button.
        btn_search = ipw.Button(
            description="Search",
            button_style="info",
            layout={"width": "initial", "margin": "2px 0 0 2em"},
        )
        btn_search.on_click(self.search)

        age_selection = ipw.VBox(
            [
                date_text,
                ipw.HBox([self.start_date_widget, self.end_date_widget, btn_search]),
            ],
            layout={"border": "1px solid #fafafa", "padding": "1em"},
        )

        h_line = ipw.HTML("<hr>")
        box = ipw.VBox([age_selection, h_line, ipw.HBox([self.mode, self.drop_label])])

        self.results = ipw.Dropdown(layout={"width": "900px"})
        self.results.observe(self._on_select_structure, names="value")
        self.search()
        super().__init__([box, h_line, self.results])

    def preprocess(self):
        """Search structures in AiiDA database and add formula extra to them."""

        queryb = orm.QueryBuilder()
        queryb.append(
            self.query_structure_type, filters={"extras": {"!has_key": "formula"}}
        )
        for item in queryb.all():  # iterall() would interfere with base.extras.set()
            try:
                formula = get_formula(item[0])
                item[0].base.extras.set("formula", formula)
            except ValueError:
                pass

    def search(self, _=None):
        """Launch the search of structures in AiiDA database."""
        self.preprocess()

        qbuild = orm.QueryBuilder()

        # If the date range is valid, use it for the search
        try:
            start_date = datetime.datetime.strptime(
                self.start_date_widget.value, "%Y-%m-%d"
            )
            end_date = datetime.datetime.strptime(
                self.end_date_widget.value, "%Y-%m-%d"
            ) + datetime.timedelta(hours=24)

        # Otherwise revert to the standard (i.e. last 7 days)
        except ValueError:
            start_date = datetime.datetime.now() - datetime.timedelta(days=7)
            end_date = datetime.datetime.now() + datetime.timedelta(hours=24)

            self.start_date_widget.value = start_date.strftime("%Y-%m-%d")
            self.end_date_widget.value = end_date.strftime("%Y-%m-%d")

        filters = {}
        filters["ctime"] = {"and": [{">": start_date}, {"<=": end_date}]}

        if self.mode.value == "uploaded":
            qbuild2 = (
                orm.QueryBuilder()
                .append(self.query_structure_type, project=["id"], tag="structures")
                .append(orm.Node, with_outgoing="structures")
            )
            processed_nodes = [n[0] for n in qbuild2.all()]
            if processed_nodes:
                filters["id"] = {"!in": processed_nodes}
            qbuild.append(self.query_structure_type, filters=filters)

        elif self.mode.value == "calculated":
            if self.drop_label.value == "All":
                qbuild.append(
                    (orm.CalcJobNode, orm.WorkChainNode), tag="calcjobworkchain"
                )
            else:
                qbuild.append(
                    (orm.CalcJobNode, orm.WorkChainNode),
                    filters={"label": self.drop_label.value},
                    tag="calcjobworkchain",
                )
            qbuild.append(
                self.query_structure_type,
                with_incoming="calcjobworkchain",
                filters=filters,
            )

        elif self.mode.value == "edited":
            qbuild.append(orm.CalcFunctionNode)
            qbuild.append(
                self.query_structure_type,
                with_incoming=orm.CalcFunctionNode,
                filters=filters,
            )

        elif self.mode.value == "all":
            qbuild.append(self.query_structure_type, filters=filters)

        qbuild.order_by({self.query_structure_type: {"ctime": "desc"}})
        matches = {n[0] for n in qbuild.iterall()}
        matches = sorted(matches, reverse=True, key=lambda n: n.ctime)

        options = [(f"Select a Structure ({len(matches)} found)", False)]
        for mch in matches:
            label = f"PK: {mch.pk}"
            label += " | " + mch.ctime.strftime("%Y-%m-%d %H:%M")
            label += " | " + mch.base.extras.get("formula", "")
            label += " | " + mch.node_type.split(".")[-2]
            label += " | " + mch.label
            label += " | " + mch.description
            options.append((label, mch))

        self.results.options = options

    def _on_select_structure(self, _=None):
        self.structure = self.results.value or None


class SmilesWidget(ipw.VBox):
    """Convert SMILES into 3D structure."""

    structure = tl.Instance(ase.Atoms, allow_none=True)

    SPINNER = """<i class="fa fa-spinner fa-pulse" style="color:red;" ></i>"""

    def __init__(self, title=""):
        self.title = title
        try:
            from rdkit import Chem  # noqa: F401
            from rdkit.Chem import AllChem  # noqa: F401
        except ImportError:
            super().__init__(
                [
                    ipw.HTML(
                        "The SmilesWidget requires the rdkit library, "
                        "but the library was not found."
                    )
                ]
            )
            return

        self.smiles = ipw.Text(placeholder="C=C")
        self.create_structure_btn = ipw.Button(
            description="Generate molecule",
            button_style="primary",
            tooltip="Generate molecule from SMILES string",
        )
        self.create_structure_btn.on_click(self._on_button_pressed)
        self.output = ipw.HTML("")

        super().__init__(
            [ipw.HBox([self.smiles, self.create_structure_btn]), self.output]
        )

    def _make_ase(self, species, positions, smiles):
        """Create ase Atoms object."""
        atoms = ase.Atoms(species, positions=positions, pbc=False)
        atoms.cell = np.ptp(atoms.positions, axis=0) + 10
        atoms.center()
        # We're attaching this info so that it
        # can be later stored as an extra on AiiDA Structure node.
        atoms.info["smiles"] = smiles

        return atoms

    def _rdkit_opt(self, smiles, steps):
        """Optimize a molecule using force field and rdkit (needed for complex SMILES)."""
        from rdkit import Chem
        from rdkit.Chem import AllChem

        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            # Something is seriously wrong with the SMILES code,
            # just return None and don't attempt anything else.
            self.output.value = "RDKit ERROR: Invalid SMILES string"
            return None
        mol = Chem.AddHs(mol)

        conf_id = AllChem.EmbedMolecule(mol, maxAttempts=20, randomSeed=42)
        if conf_id < 0:
            # Retry with different generation method that is supposed to be
            # more stable. Perhaps we should switch to it by default.
            # https://greglandrum.github.io/rdkit-blog/posts/2021-01-31-looking-at-random-coordinate-embedding.html#look-at-some-of-the-troublesome-structures
            # https://www.rdkit.org/docs/source/rdkit.Chem.rdDistGeom.html#rdkit.Chem.rdDistGeom.EmbedMolecule
            conf_id = AllChem.EmbedMolecule(
                mol, maxAttempts=20, useRandomCoords=True, randomSeed=422
            )
        if conf_id < 0:
            self.output.value = "RDKit ERROR: Could not generate conformer"
            return None
        if AllChem.UFFHasAllMoleculeParams(mol):
            AllChem.UFFOptimizeMolecule(mol, maxIters=steps)
        else:
            self.output.value = "RDKit WARNING: Missing UFF parameters"

        positions = mol.GetConformer().GetPositions()
        natoms = mol.GetNumAtoms()
        species = [mol.GetAtomWithIdx(j).GetSymbol() for j in range(natoms)]
        return self._make_ase(species, positions, smiles)

    def _mol_from_smiles(self, smiles, steps=1000):
        """Convert SMILES to ASE structure using RDKit"""
        try:
            canonical_smiles = self.canonicalize_smiles(smiles)
            ase = self._rdkit_opt(canonical_smiles, steps)
        except ValueError as e:
            self.output.value = str(e)
            return None
        else:
            if canonical_smiles != smiles:
                self.output.value = f"Canonical SMILES: {canonical_smiles}"
            return ase

    def _on_button_pressed(self, change=None):
        """Convert SMILES to ASE structure when button is pressed."""
        self.output.value = ""

        if not self.smiles.value:
            return
        spinner = f"Screening possible conformers {self.SPINNER}"  # font-size:20em;
        self.output.value = spinner

        self.structure = self._mol_from_smiles(self.smiles.value)
        # Don't overwrite possible error/warning messages
        if self.output.value == spinner:
            self.output.value = ""

    # https://en.wikipedia.org/wiki/Simplified_molecular-input_line-entry_system#Terminology
    @staticmethod
    def canonicalize_smiles(smiles: str) -> str:
        """Canonicalize the SMILES code.

        :raises ValueError: if SMILES is invalid or if canonicalization fails
        """
        from rdkit import Chem

        mol = Chem.MolFromSmiles(smiles, sanitize=True)
        if mol is None:
            # Something is seriously wrong with the SMILES code
            msg = "Invalid SMILES string"
            raise ValueError(msg)

        canonical_smiles = Chem.MolToSmiles(mol, isomericSmiles=True, canonical=True)
        if not canonical_smiles:
            msg = "SMILES canonicalization failed"
            raise ValueError(msg)
        return canonical_smiles

    @tl.default("structure")
    def _default_structure(self):
        return None


def _register_structure(operator):
    """
    Decorator for methods that manipulate (operate on) the selected structure.

    Checks whether a structure and selection is set and ensures that the
    arguments for structure and selection are passed by copy rather than
    reference. A pop-up warning message is shown in case that neither a
    structure or selection are set.
    """

    @functools.wraps(operator)
    def inner(ref, *args, **kwargs):
        if not ref.structure:
            ref._status_message.message = """
            <div class="alert alert-info">
            <strong>Please choose a structure first.</strong>
            </div>
            """
        else:
            operator(
                ref,
                *args,
                **kwargs,
                atoms=ref.structure.copy(),
            )

    return inner


def _register_selection(operator):
    """
    Decorator for methods that manipulate (operate on) the selected structure.

    Checks whether a structure and selection is set and ensures that the
    arguments for structure and selection are passed by copy rather than
    reference. A pop-up warning message is shown in case that neither a
    structure or selection are set.
    """

    @functools.wraps(operator)
    def inner(ref, *args, **kwargs):
        if not ref.selection:
            ref._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select atoms first.</strong>
            </div>
            """
        else:
            operator(
                ref,
                *args,
                **kwargs,
                selection=ref.selection.copy(),
            )

    return inner


class BasicCellEditor(ipw.VBox):
    """Widget that allows for the basic cell editing."""

    structure = tl.Instance(ase.Atoms, allow_none=True)

    def __init__(self, title="Cell transform"):
        self.title = title
        self._status_message = StatusHTML()

        # cell transfor opration widget
        primitive_cell = ipw.Button(
            description="Convert to primitive cell",
            layout={"width": "initial"},
        )
        primitive_cell.on_click(self._to_primitive_cell)

        conventional_cell = ipw.Button(
            description="Convert to conventional cell",
            layout={"width": "initial"},
        )
        conventional_cell.on_click(self._to_conventional_cell)
        cell_parameters = (
            self.structure.get_cell_lengths_and_angles()
            if self.structure
            else [1, 0, 0, 0, 0, 0]
        )
        self.cell_parameters = ipw.HBox(
            [
                ipw.VBox(
                    [
                        ipw.HTML(
                            description=["a(Å)", "b(Å)", "c(Å)", "α", "β", "γ"][i],
                            layout={"width": "30px"},
                        ),
                        ipw.FloatText(
                            value=cell_parameters[i], layout={"width": "100px"}
                        ),
                    ]
                )
                for i in range(6)
            ]
        )
        # cell transformation matrix (4x4)
        self.cell_transformation = ipw.VBox(
            [
                ipw.HBox(
                    [
                        ipw.IntText(value=1 if i == j else 0, layout={"width": "60px"})
                        for j in range(3)
                    ]
                    + [ipw.FloatText(value=0, layout={"width": "60px"})]
                )
                for i in range(3)
            ]
        )
        apply_cell_parameters_button = ipw.Button(description="Apply cell parameters")
        apply_cell_parameters_button.on_click(self._apply_cell_parameters)
        self.scale_atoms_position = ipw.Checkbox(
            description="Scale atoms position",
            value=False,
            indent=False,
        )
        apply_cell_transformation = ipw.Button(description="Apply transformation")
        apply_cell_transformation.on_click(self._apply_cell_transformation)
        reset_transformatioin_button = ipw.Button(
            description="Reset matrix",
        )
        reset_transformatioin_button.on_click(self._reset_cell_transformation_matrix)
        super().__init__(
            children=[
                ipw.HBox(
                    [
                        primitive_cell,
                        conventional_cell,
                    ],
                ),
                self._status_message,
                ipw.VBox(
                    [
                        ipw.HTML(
                            "<b>Cell parameters:</b>",
                            layout={"margin": "20px 0px 10px 0px"},
                        ),
                        self.cell_parameters,
                        ipw.HBox(
                            [
                                apply_cell_parameters_button,
                                self.scale_atoms_position,
                            ]
                        ),
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.VBox(
                    [
                        ipw.HTML(
                            "<b>Cell transformation:</b>",
                            layout={"margin": "20px 0px 10px 0px"},
                        ),
                        self.cell_transformation,
                        ipw.HBox(
                            [apply_cell_transformation, reset_transformatioin_button]
                        ),
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
            ],
        )

    @_register_structure
    def _to_primitive_cell(self, _=None, atoms=None):
        atoms = self._to_standard_cell(atoms, to_primitive=True)

        self.structure = atoms

    @_register_structure
    def _to_conventional_cell(self, _=None, atoms=None):
        atoms = self._to_standard_cell(atoms, to_primitive=False)

        self.structure = atoms

    @staticmethod
    def _to_standard_cell(
        structure: ase.Atoms, to_primitive=False, no_idealize=False, symprec=1e-5
    ):
        """The `standardize_cell` method from spglib and apply to ase.Atoms"""
        lattice = structure.get_cell()
        positions = structure.get_scaled_positions()
        numbers = structure.get_atomic_numbers()

        cell = (lattice, positions, numbers)

        # operation
        lattice, positions, numbers = spglib.standardize_cell(
            cell, to_primitive=to_primitive, no_idealize=no_idealize, symprec=symprec
        )

        return ase.Atoms(
            cell=lattice,
            scaled_positions=positions,
            numbers=numbers,
            pbc=[True, True, True],
        )

    @tl.observe("structure")
    def _observe_structure(self, change):
        """Update cell after the structure has been modified."""
        if change["new"] is not None:
            cell_parameters = change["new"].cell.cellpar()
            for i in range(6):
                self.cell_parameters.children[i].children[1].value = round(
                    cell_parameters[i], 4
                )
        else:
            for i in range(6):
                self.cell_parameters.children[i].children[1].value = 0

    @_register_structure
    def _apply_cell_parameters(self, _=None, atoms=None):
        """Apply the cell parameters to the structure."""
        # only update structure when atoms is not None.
        cell_parameters = [
            self.cell_parameters.children[i].children[1].value for i in range(6)
        ]
        if atoms is not None:
            atoms.set_cell(
                ase.cell.Cell.fromcellpar(cell_parameters),
                scale_atoms=self.scale_atoms_position.value,
            )
            self.structure = atoms

    @_register_structure
    def _apply_cell_transformation(self, _=None, atoms=None):
        """Apply the transformation matrix to the structure."""
        from ase.build import make_supercell

        # only update structure when atoms is not None.
        if atoms is not None:
            mat = np.zeros((3, 3))
            translate = np.zeros(3)
            for i in range(3):
                translate[i] = self.cell_transformation.children[i].children[3].value
                for j in range(3):
                    mat[i][j] = self.cell_transformation.children[i].children[j].value
            # transformation matrix may not work due to singularity, or
            # the number of generated atoms is not correct
            try:
                atoms = make_supercell(atoms, mat)
            except Exception as e:
                self._status_message.message = f"""
            <div class="alert alert-info">
            <strong>The transformation matrix is wrong! {e}</strong>
            </div>
            """
                return
            # translate
            atoms.translate(-atoms.cell.array.dot(translate))
            self.structure = atoms

    @_register_structure
    def _reset_cell_transformation_matrix(self, _=None, atoms=None):
        """Reset the transformation matrix to identity matrix."""
        for i in range(3):
            for j in range(4):
                self.cell_transformation.children[i].children[j].value = 0
            self.cell_transformation.children[i].children[i].value = 1


class BasicStructureEditor(ipw.VBox):
    """
    Widget that allows for the basic structure (molecule and
    position of periodic structure in cell) editing."""

    structure = tl.Instance(ase.Atoms, allow_none=True)
    input_selection = tl.List(tl.Int(), allow_none=True)
    selection = tl.List(tl.Int())
    camera_orientation = tl.List()

    def __init__(self, title=""):
        self.title = title

        # Define action vector.
        self.axis_p1 = ipw.Text(
            description="Starting point", value="0 0 0", layout={"width": "initial"}
        )
        self.axis_p2 = ipw.Text(
            description="Ending point", value="0 0 1", layout={"width": "initial"}
        )
        btn_def_atom1 = ipw.Button(
            description="From selection", layout={"width": "initial"}
        )
        btn_def_atom1.on_click(self.def_axis_p1)
        btn_def_atom2 = ipw.Button(
            description="From selection", layout={"width": "initial"}
        )
        btn_def_atom2.on_click(self.def_axis_p2)
        btn_get_from_camera = ipw.Button(
            description="Perp. to screen",
            button_style="warning",
            layout={"width": "initial"},
        )
        btn_get_from_camera.on_click(self.def_perpendicular_to_screen)

        # Define action point.
        self.point = ipw.Text(
            description="Action point", value="0 0 0", layout={"width": "initial"}
        )
        btn_def_pnt = ipw.Button(
            description="From selection", layout={"width": "initial"}
        )
        btn_def_pnt.on_click(self.def_point)

        # Move atoms.
        btn_move_dr = ipw.Button(description="Move", layout={"width": "initial"})
        btn_move_dr.on_click(self.translate_dr)
        self.displacement = ipw.FloatText(
            description=r"Move along action vector by $\Delta=$ ",
            value=1,
            step=0.1,
            style={"description_width": "initial"},
            layout={"width": "initial"},
        )

        btn_move_dxyz = ipw.Button(
            description="Move by XYZ", layout={"width": "initial"}
        )
        btn_move_dxyz.on_click(self.translate_dxdydz)
        btn_move_to_xyz = ipw.Button(
            description="Move to XYZ", layout={"width": "initial"}
        )
        btn_move_to_xyz.on_click(self.translate_to_xyz)
        self.dxyz = ipw.Text(
            description="XYZ move:",
            value="0 0 0",
            style={"description_width": "initial"},
            layout={"width": "initial", "margin": "0px 0px 0px 20px"},
        )

        # Rotate atoms.
        btn_rotate = ipw.Button(description="Rotate", layout={"width": "10%"})
        btn_rotate.on_click(self.rotate)
        self.phi = ipw.FloatText(
            description="Rotate around the action vector which starts from the action point",
            value=0,
            step=5,
            style={"description_width": "initial"},
            layout={"width": "initial"},
        )

        # Mirror atoms.
        btn_mirror_perp = ipw.Button(description="Mirror", layout={"width": "10%"})
        btn_mirror_perp.on_click(self.mirror)
        btn_mirror_3p = ipw.Button(description="Mirror", layout={"width": "10%"})
        btn_mirror_3p.on_click(self.mirror_3p)

        # Rotate atoms while aligning action vector with XYZ vector.
        btn_align = ipw.Button(description="Align", layout={"width": "10%"})
        btn_align.on_click(self.align)

        # Atoms selection.
        self.element = ipw.Dropdown(
            description="Select element",
            options=ase.data.chemical_symbols[1:],
            value="H",
            style={"description_width": "initial"},
            layout={"width": "initial"},
        )

        def disable_element(_=None):
            if self.ligand.value == 0:
                self.element.disabled = False
            else:
                self.element.disabled = True

        # Ligand selection.
        self.ligand = FunctionalGroupSelectorWidget()
        self.ligand.observe(disable_element, names="value")

        # Add atom.
        btn_add = ipw.Button(description="Add to selected", layout={"width": "initial"})
        btn_add.on_click(self.add)
        self.bond_length = ipw.FloatText(
            description="Bond length", value=1.0, layout={"width": "140px"}
        )
        use_covalent_radius = ipw.Checkbox(
            value=True,
            description="Use covalent radius",
            style={"description_width": "initial"},
        )
        tl.link((use_covalent_radius, "value"), (self.bond_length, "disabled"))

        # Copy atoms.
        btn_copy_sel = ipw.Button(
            description="Copy selected", layout={"width": "initial"}
        )
        btn_copy_sel.on_click(self.copy_sel)

        # Modify atom.
        btn_modify = ipw.Button(
            description="Modify selected",
            button_style="warning",
            layout={"width": "initial"},
        )
        btn_modify.on_click(self.mod_element)

        # Remove atom.
        btn_remove = ipw.Button(
            description="Remove selected",
            button_style="danger",
            layout={"width": "initial"},
        )
        btn_remove.on_click(self.remove)

        # Automatically clear selection after point definition
        self.autoclear_selection = ipw.Checkbox(
            description='Clear selection after pressing "From seletion"',
            value=True,
            style={"description_width": "initial"},
        )

        self._status_message = StatusHTML()

        super().__init__(
            children=[
                ipw.HTML(
                    "<b>Action vector and point:</b>",
                    layout={"margin": "20px 0px 10px 0px"},
                ),
                ipw.HBox(
                    [
                        self.axis_p1,
                        btn_def_atom1,
                        self.axis_p2,
                        btn_def_atom2,
                        btn_get_from_camera,
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HBox(
                    [self.point, btn_def_pnt, self.autoclear_selection],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HTML(
                    "<b>Move atom(s):</b>", layout={"margin": "20px 0px 10px 0px"}
                ),
                ipw.HBox(
                    [
                        self.displacement,
                        btn_move_dr,
                        self.dxyz,
                        btn_move_dxyz,
                        btn_move_to_xyz,
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HBox([self.phi, btn_rotate], layout={"margin": "0px 0px 0px 20px"}),
                ipw.HBox(
                    [
                        ipw.HTML(
                            "Mirror on the plane perpendicular to the action vector",
                            layout={"margin": "0px 0px 0px 0px"},
                        ),
                        btn_mirror_perp,
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HBox(
                    [
                        ipw.HTML(
                            "Mirror on the plane containing action vector and action point",
                            layout={"margin": "0px 0px 0px 0px"},
                        ),
                        btn_mirror_3p,
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HBox(
                    [
                        ipw.HTML(
                            "Rotate atoms while aligning the action vector with the XYZ vector",
                            layout={"margin": "0px 0px 0px 0px"},
                        ),
                        btn_align,
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HTML(
                    "<b>Modify atom(s):</v>", layout={"margin": "20px 0px 10px 0px"}
                ),
                ipw.HBox([btn_copy_sel], layout={"margin": "0px 0px 0px 20px"}),
                ipw.HBox(
                    [self.element, self.ligand], layout={"margin": "0px 0px 0px 20px"}
                ),
                ipw.HBox(
                    [
                        btn_modify,
                        btn_add,
                        self.bond_length,
                        use_covalent_radius,
                    ],
                    layout={"margin": "0px 0px 0px 20px"},
                ),
                ipw.HBox([btn_remove], layout={"margin": "0px 0px 0px 20px"}),
                self._status_message,
            ]
        )

    def str2vec(self, string):
        return np.array(list(map(float, string.split())))

    def vec2str(self, vector):
        return (
            str(round(vector[0], 2))
            + " "
            + str(round(vector[1], 2))
            + " "
            + str(round(vector[2], 2))
        )

    def sel2com(self):
        """Get center of mass of the selection."""
        if self.selection:
            com = np.average(self.structure[self.selection].get_positions(), axis=0)
        else:
            com = [0, 0, 0]

        return com

    @property
    def action_vector(self):
        """Define the action vector."""
        normal = self.str2vec(self.axis_p2.value) - self.str2vec(self.axis_p1.value)
        return normal / np.linalg.norm(normal)

    def def_point(self, _=None):
        """Define the action point."""
        self.point.value = self.vec2str(self.sel2com())
        if self.autoclear_selection.value:
            self.input_selection = None
            self.input_selection = []

    def def_axis_p1(self, _=None):
        """Define the first point of axis."""
        self.axis_p1.value = self.vec2str(self.sel2com())
        if self.autoclear_selection.value:
            self.input_selection = None
            self.input_selection = []

    def def_axis_p2(self, _=None):
        """Define the second point of axis."""
        if not self.selection:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select atoms first.</strong>
            </div>
            """
        else:
            com = (
                np.average(self.structure[self.selection].get_positions(), axis=0)
                if self.selection
                else [0, 0, 1]
            )
            self.axis_p2.value = self.vec2str(com)
            if self.autoclear_selection.value:
                self.input_selection = None
                self.input_selection = []

    def def_perpendicular_to_screen(self, _=None):
        """Define a normalized vector perpendicular to the screen."""
        cmr = self.camera_orientation
        if cmr:
            self.axis_p1.value = "0 0 0"
            versor = np.array([-cmr[2], -cmr[6], -cmr[10]]) / np.linalg.norm(
                [-cmr[2], -cmr[6], -cmr[10]]
            )
            self.axis_p2.value = self.vec2str(versor.tolist())

    @_register_structure
    @_register_selection
    def translate_dr(self, _=None, atoms=None, selection=None):
        """Translate by dr along the selected vector."""

        atoms.positions[self.selection] += np.array(
            self.action_vector * self.displacement.value
        )

        self.input_selection = None  # Clear selection.

        self.structure, self.input_selection = atoms, selection

    @_register_structure
    @_register_selection
    def translate_dxdydz(self, _=None, atoms=None, selection=None):
        """Translate by the selected XYZ delta."""

        # The action.
        atoms.positions[self.selection] += np.array(self.str2vec(self.dxyz.value))
        self.input_selection = None  # Clear selection.
        self.structure, self.input_selection = atoms, selection

    @_register_structure
    @_register_selection
    def translate_to_xyz(self, _=None, atoms=None, selection=None):
        """Translate to the selected XYZ position."""
        # The action.
        geo_center = np.average(self.structure[self.selection].get_positions(), axis=0)
        atoms.positions[self.selection] += self.str2vec(self.dxyz.value) - geo_center
        self.input_selection = None  # Clear selection.
        self.structure, self.input_selection = atoms, selection

    @_register_structure
    @_register_selection
    def rotate(self, _=None, atoms=None, selection=None):
        """Rotate atoms around selected point in space and vector."""

        # The action.
        rotated_subset = atoms[self.selection]
        vec = self.str2vec(self.vec2str(self.action_vector))
        center = self.str2vec(self.point.value)
        rotated_subset.rotate(self.phi.value, v=vec, center=center, rotate_cell=False)
        atoms.positions[self.selection] = rotated_subset.positions
        self.input_selection = None  # Clear selection.

        self.structure, self.input_selection = atoms, selection

    @_register_structure
    @_register_selection
    def mirror(self, _=None, norm=None, point=None, atoms=None, selection=None):
        """Mirror atoms on the plane perpendicular to the action vector."""
        # The action.
        # Vector and point define the mirrorring surface.
        p_normal = norm if norm is not None else self.action_vector
        p_point = point if point is not None else self.str2vec(self.point.value)

        # Check if norm vector makes sense.
        if np.isnan(p_normal).any() or np.linalg.norm(p_normal) < 1e-4:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Norm vector not makes sense.</strong>
            </div>
            """
            return

        # Define vectors from p_point that point to the atoms which are to be moved.
        mirror_subset = atoms.positions[selection] - p_point

        # Project vectors onto the plane normal.
        projections = (
            p_normal
            * np.dot(mirror_subset, p_normal)[:, np.newaxis]
            / np.dot(p_normal, p_normal)
        )

        # Mirror atoms.
        atoms.positions[selection] -= 2 * projections

        self.input_selection = None  # Clear selection.

        self.structure, self.input_selection = atoms, selection

    def mirror_3p(self, _=None):
        """Mirror atoms on the plane containing action vector and action point."""
        pt1 = self.str2vec(self.axis_p2.value)
        pt2 = self.str2vec(self.axis_p1.value)
        pt3 = self.str2vec(self.point.value)
        normal = np.cross(pt1 - pt2, pt3 - pt2)
        normal = normal / np.linalg.norm(normal)
        self.mirror(norm=normal, point=pt3)

    @_register_structure
    @_register_selection
    def align(self, _=None, atoms=None, selection=None):
        """Rotate atoms to align action vector with XYZ vector."""
        if not self.selection:
            return

        # The action.
        center = self.str2vec(self.point.value)
        subset = atoms[selection]
        subset.rotate(self.action_vector, self.str2vec(self.dxyz.value), center=center)
        atoms.positions[selection] = subset.positions

        self.structure, self.input_selection = atoms, selection

    @_register_structure
    @_register_selection
    def mod_element(self, _=None, atoms=None, selection=None):
        """Modify selected atoms into the given element."""
        last_atom = atoms.get_global_number_of_atoms()

        if self.ligand.value == 0:
            for idx in self.selection:
                new = ase.Atom(self.element.value)
                atoms[idx].mass = new.mass
                atoms[idx].magmom = new.magmom
                atoms[idx].momentum = new.momentum
                atoms[idx].symbol = new.symbol
                atoms[idx].tag = new.tag
                atoms[idx].charge = new.charge
            new_selection = selection
        else:
            initial_ligand = self.ligand.rotate(
                align_to=self.action_vector, remove_anchor=True
            )

            for idx in self.selection:
                position = self.structure.positions[idx].copy()
                lgnd = initial_ligand.copy()
                lgnd.translate(position)
                atoms += lgnd
            new_selection = list(
                range(last_atom, last_atom + len(selection) * len(lgnd))
            )

        self.structure, self.input_selection = atoms, new_selection

    @_register_structure
    @_register_selection
    def copy_sel(self, _=None, atoms=None, selection=None):
        """Copy selected atoms and shift by 1.0 A along X-axis."""

        last_atom = atoms.get_global_number_of_atoms()

        # The action
        add_atoms = atoms[self.selection].copy()
        add_atoms.translate([1.0, 0, 0])
        atoms += add_atoms

        self.structure, self.input_selection = (
            atoms,
            list(range(last_atom, last_atom + len(selection))),
        )

    @_register_structure
    @_register_selection
    def add(self, _=None, atoms=None, selection=None):
        """Add atoms."""
        last_atom = atoms.get_global_number_of_atoms()
        if self.ligand.value == 0:
            initial_ligand = ase.Atoms([ase.Atom(self.element.value, [0, 0, 0])])
            rad = SYMBOL_RADIUS[self.element.value]
        else:
            initial_ligand = self.ligand.rotate(align_to=self.action_vector)
            rad = SYMBOL_RADIUS[self.ligand.anchoring_atom]

        for idx in self.selection:
            # It is important to copy, otherwise the initial structure will be modified
            position = self.structure.positions[idx].copy()
            lgnd = initial_ligand.copy()

            if self.bond_length.disabled:
                lgnd.translate(
                    position
                    + self.action_vector
                    * (SYMBOL_RADIUS[self.structure.symbols[idx]] + rad)
                )
            else:
                lgnd.translate(position + self.action_vector * self.bond_length.value)

            atoms += lgnd

        new_selection = list(range(last_atom, last_atom + len(selection) * len(lgnd)))

        # The order of the traitlets below is important -
        # we must be sure trait atoms is set before trait selection
        self.structure, self.input_selection = atoms, new_selection

    @_register_structure
    @_register_selection
    def remove(self, _=None, atoms=None, selection=None):
        """Remove selected atoms."""
        del [atoms[selection]]

        # The order of the traitlets below is important -
        # we must be sure trait atoms is set before trait selection
        self.structure = atoms
        self.input_selection = None



"""Some utility functions used acrross the repository."""

import itertools
import operator
import threading
from enum import Enum
from typing import Any

import ase
import ase.io
import ipywidgets as ipw
import numpy as np
import traitlets as tl
from aiida.plugins import DataFactory

CifData = DataFactory("core.cif")  # pylint: disable=invalid-name
StructureData = DataFactory("core.structure")  # pylint: disable=invalid-name
TrajectoryData = DataFactory("core.array.trajectory")  # pylint: disable=invalid-name


def valid_arguments(arguments, valid_args):
    """Check whether provided arguments are valid."""
    result = {}
    for key, value in arguments.items():
        if key in valid_args:
            if isinstance(value, (tuple, list)):
                result[key] = "\n".join(value)
            else:
                result[key] = value
    return result


def predefine_settings(obj, **kwargs):
    """Specify some pre-defined settings."""
    for key, value in kwargs.items():
        if hasattr(obj, key):
            setattr(obj, key, value)
        else:
            raise AttributeError(f"{obj!r} object has no attribute {key!r}")


def get_ase_from_file(fname, file_format=None):  # pylint: disable=redefined-builtin
    """Get ASE structure object."""
    # store_tags parameter is useful for CIF files
    # https://wiki.fysik.dtu.dk/ase/ase/io/formatoptions.html#cif
    if file_format == "cif":
        traj = ase.io.read(fname, format=file_format, index=":", store_tags=True)
    else:
        traj = ase.io.read(fname, format=file_format, index=":")
    if not traj:
        raise ValueError(f"Could not read any information from the file {fname}")
    return traj


def find_ranges(iterable):
    """Yield range of consecutive numbers."""
    for grp in _consecutive_groups(iterable):
        group = list(grp)
        if len(group) == 1:
            yield group[0]
        else:
            yield group[0], group[-1]


def _consecutive_groups(iterable, ordering=lambda x: x):
    """Yield groups of consecutive items using :func:`itertools.groupby`.
    The *ordering* function determines whether two items are adjacent by
    returning their position.

    This is a vendored version of more_itertools.consecutive_groups
    https://more-itertools.readthedocs.io/en/v10.3.0/_modules/more_itertools/more.html#consecutive_groups
    Distributed under MIT license: https://more-itertools.readthedocs.io/en/v10.3.0/license.html
    Thank you Bo Bayles for the original implementation. <3
    """
    for _, g in itertools.groupby(
        enumerate(iterable), key=lambda x: x[0] - ordering(x[1])
    ):
        yield map(operator.itemgetter(1), g)


def list_to_string_range(lst, shift=1):
    """Converts a list like [0, 2, 3, 4] into a string like '1 3..5'.

    Shift used when e.g. for a user interface numbering starts from 1 not from 0"""
    return " ".join(
        [
            f"{t[0] + shift}..{t[1] + shift}"
            if isinstance(t, tuple)
            else str(t + shift)
            for t in find_ranges(sorted(lst))
        ]
    )


def string_range_to_list(strng, shift=-1):
    """Converts a string like '1 3..5' into a list like [0, 2, 3, 4].

    Shift used when e.g. for a user interface numbering starts from 1 not from 0"""
    singles = [int(s) + shift for s in strng.split() if s.isdigit()]
    ranges = [r for r in strng.split() if ".." in r]
    if len(singles) + len(ranges) != len(strng.split()):
        return [], False
    for rng in ranges:
        try:
            start, end = rng.split("..")
            singles += [i + shift for i in range(int(start), int(end) + 1)]
        except ValueError:
            return [], False
    return singles, True


def get_formula(data_node):
    """A wrapper for getting a molecular formula out of the AiiDA Data node"""
    if isinstance(data_node, TrajectoryData):
        # TrajectoryData can only hold structures with the same chemical formula,
        # so this approach is sound.
        stepid = data_node.get_stepids()[0]
        return data_node.get_step_structure(stepid).get_formula()
    elif isinstance(data_node, StructureData):
        return data_node.get_formula()
    elif isinstance(data_node, CifData):
        return data_node.get_ase().get_chemical_formula()
    else:
        raise TypeError(f"Cannot get formula from node {type(data_node)}")


class PinholeCamera:
    def __init__(self, matrix):
        self.matrix = np.reshape(matrix, (4, 4)).transpose()

    def screen_to_vector(self, move_vector):
        """Converts vector from the screen coordinates to the normalized vector in 3D."""
        move_vector[0] = -move_vector[0]  # the x axis seem to be reverted in nglview.
        res = np.append(np.array(move_vector), [0])
        res = self.inverse_matrix.dot(res)
        res /= np.linalg.norm(res)
        return res[0:3]

    @property
    def inverse_matrix(self):
        return np.linalg.inv(self.matrix)


class _StatusWidgetMixin(tl.HasTraits):
    """Show temporary messages for example for status updates.
    This is a mixin class that is meant to be part of an inheritance
    tree of an actual widget with a 'value' traitlet that is used
    to convey a status message. See the non-private classes below
    for examples.
    """

    message = tl.Unicode(default_value="", allow_none=True)
    new_line = "\n"

    def __init__(self, clear_after=3, *args, **kwargs):
        self._clear_timer = None
        self._clear_after = clear_after
        self._message_stack = []
        super().__init__(*args, **kwargs)

    def _clear_value(self):
        """Set widget .value to be an empty string."""
        if self._message_stack:
            self._message_stack.pop(0)
            self.value = self.new_line.join(self._message_stack)
        else:
            self.value = ""

    def show_temporary_message(self, value, clear_after=None):
        """Show a temporary message and clear it after the given interval."""
        clear_after = clear_after or self._clear_after
        if value:
            self._message_stack.append(value)
            self.value = self.new_line.join(self._message_stack)

            # Start new timer that will clear the value after the specified interval.
            self._clear_timer = threading.Timer(self._clear_after, self._clear_value)
            self._clear_timer.start()
            self.message = None


class StatusHTML(_StatusWidgetMixin, ipw.HTML):
    """Show temporary HTML messages for example for status updates."""

    new_line = "<br>"

    # This method should be part of _StatusWidgetMixin, but that does not work
    # for an unknown reason.
    @tl.observe("message")
    def _observe_message(self, change):
        self.show_temporary_message(change["new"])


# Define the message levels as Enum
class MessageLevel(Enum):
    INFO = "info"
    WARNING = "warning"
    ERROR = "danger"
    SUCCESS = "success"


def wrap_message(message, level=MessageLevel.INFO):
    """Wrap message into HTML code with the given level."""
    # mapping level to fa icon
    # https://fontawesome.com/v4.7.0/icons/
    mapping = {
        MessageLevel.INFO: "info-circle",
        MessageLevel.WARNING: "exclamation-triangle",
        MessageLevel.ERROR: "exclamation-circle",
        MessageLevel.SUCCESS: "check-circle",
    }

    # The message is wrapped into a div with the class "alert" and the icon of the given level
    return f"""
        <div class="alert alert-{level.value}" role="alert" style="margin-bottom: 0px; padding: 6px 12px;">
            <i class="fa fa-{mapping[level]}"></i> {message}
        </div>
    """


def ase2spglib(ase_structure: ase.Atoms) -> tuple[Any, Any, Any]:
    """
    Convert ase Atoms instance to spglib cell in the format defined at
    https://spglib.github.io/spglib/python-spglib.html#crystal-structure-cell
    """
    lattice = ase_structure.get_cell()
    positions = ase_structure.get_scaled_positions()
    numbers = ase_structure.get_atomic_numbers()

    return (lattice, positions, numbers)



import base64
import json
import zlib

from aiidalab_widgets_base import bug_report


def test_fingerprint_parser():
    """Test get_environment_fingerprint function and parse it out."""

    encoding = "utf-8"
    fingerprint = bug_report.get_environment_fingerprint(encoding)

    # Parse the fingerprint.
    data = bug_report.parse_environment_fingerprint(fingerprint)

    # To test, manually generate the fingerprint and compare it to the output of the parser.
    json_data = json.dumps(data, separators=(",", ":"))
    got = base64.urlsafe_b64encode(zlib.compress(json_data.encode(encoding), level=9))

    assert got == fingerprint



from pathlib import Path

import ase
import numpy as np
import pytest

import aiidalab_widgets_base as awb


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_manager_widget_different_number_of_importers_or_editors(
    structure_data_object,
):
    """Test the `StructureManagerWidget` with different number of importers or editors."""

    # No importers or editors
    structure_manager_widget = awb.StructureManagerWidget(
        importers=[],
        editors=[],
        input_structure=None,
    )
    assert structure_manager_widget.input_structure is None
    structure_manager_widget.input_structure = structure_data_object
    assert isinstance(structure_manager_widget.structure, ase.Atoms)

    # One importer, no editors
    upload_from_computer = awb.StructureUploadWidget(title="From computer")
    structure_manager_widget = awb.StructureManagerWidget(
        importers=[
            upload_from_computer,
        ],
        editors=[],
        input_structure=None,
    )
    upload_from_computer.structure = structure_data_object
    assert structure_manager_widget.input_structure.uuid == structure_data_object.uuid

    # No importers, one editor
    structure_manager_widget = awb.StructureManagerWidget(
        importers=[],
        editors=[
            awb.BasicStructureEditor(title="Basic Editor"),
        ],
        input_structure=structure_data_object,
    )
    assert structure_manager_widget.structure is not None

    # Several importers and editors.
    base_editor = awb.BasicStructureEditor(title="Basic Editor")
    structure_manager_widget = awb.StructureManagerWidget(
        importers=[
            awb.StructureUploadWidget(title="From computer"),
            awb.StructureBrowserWidget(title="AiiDA database"),
        ],
        editors=[
            base_editor,
            awb.BasicCellEditor(title="Cell Editor"),
        ],
        viewer=awb.viewers.StructureDataViewer(),
        node_class="StructureData",
    )

    assert structure_manager_widget.structure is None
    structure_manager_widget.input_structure = structure_data_object


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_manager_widget(structure_data_object):
    """Test the `StructureManagerWidget`."""
    structure_manager_widget = awb.StructureManagerWidget(
        importers=[
            awb.StructureUploadWidget(title="From computer"),
        ],
        editors=[
            awb.BasicStructureEditor(title="Basic Editor"),
        ],
        input_structure=structure_data_object,
    )

    assert structure_manager_widget.structure is not None
    assert isinstance(structure_manager_widget.structure, ase.Atoms)

    # Test the structure periodicity
    assert structure_manager_widget.viewer.periodicity.value == "Periodicity: xyz"

    # Store structure and check that it is stored.
    structure_manager_widget.btn_store.click()
    assert structure_manager_widget.structure_node.is_stored
    assert structure_manager_widget.structure_node.pk is not None

    # Try to store the stored structure.
    structure_manager_widget.btn_store.click()
    assert "Already stored" in structure_manager_widget.output.value

    # Simulate the structure modification.
    new_structure = structure_manager_widget.structure.copy()
    new_structure[0].position += [0, 0, 1]

    structure_manager_widget.structure = new_structure
    assert not structure_manager_widget.structure_node.is_stored
    assert np.all(
        structure_manager_widget.structure[0].position == new_structure[0].position
    )

    # Store the modified structure.
    structure_manager_widget.btn_store.click()
    assert structure_manager_widget.structure_node.is_stored

    # Undo the structure modification.
    structure_manager_widget.undo()
    assert np.any(
        structure_manager_widget.structure[0].position != new_structure[0].position
    )
    structure_manager_widget.undo()  # Undo the structure creation.
    assert structure_manager_widget.structure is None


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_browser_widget(structure_data_object):
    """Test the `StructureBrowserWidget`."""
    structure_browser_widget = awb.StructureBrowserWidget()
    assert structure_browser_widget.structure is None

    structure_browser_widget.search()

    # There should be no structures in the database, only the default output "Select structure".
    assert len(structure_browser_widget.results.options) == 1

    # Store the structure and check that it is listed in the widget.
    structure_data_object.store()
    structure_browser_widget.search()
    assert len(structure_browser_widget.results.options) == 2

    # Simulate the structure selection.
    structure_browser_widget.results.value = structure_data_object

    assert structure_browser_widget.structure.uuid == structure_data_object.uuid


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_upload_widget():
    """Test the `StructureUploadWidget`."""
    widget = awb.StructureUploadWidget()
    assert widget.structure is None

    # Simulate the structure upload.
    widget._on_file_upload(
        change={
            "new": {
                "test.xyz": {
                    "content": b"""2

                Si 0.0 0.0 0.0
                Si 0.5 0.5 0.5
                """,
                }
            }
        }
    )
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "Si2"
    assert np.all(widget.structure[0].position == [0, 0, 0])


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_examples_widget():
    """Test the `StructureExamplesWidget`."""
    this_folder = Path(__file__).parent

    widget = awb.StructureExamplesWidget(
        examples=[
            (
                "Silicon oxide",
                this_folder / ".." / "miscellaneous" / "structures" / "SiO2.xyz",
            )
        ]
    )
    assert widget.structure is None

    # Simulate the structure selection.
    widget._select_structure.label = "Silicon oxide"
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "O4Si2"


@pytest.mark.usefixtures("aiida_profile_clean")
def test_smiles_widget():
    """Test the `SmilesWidget`."""
    widget = awb.SmilesWidget()
    assert widget.structure is None

    # Simulate the structure generation.
    widget.smiles.value = "C"
    widget._on_button_pressed()
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "CH4"

    # Regression test that we can generate 1-atom and 2-atom molecules
    widget.smiles.value = "[O]"
    widget._on_button_pressed()
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "O"

    widget.smiles.value = "N#N"
    widget._on_button_pressed()
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "N2"

    # Should not raise for invalid smiles
    widget.smiles.value = "invalid"
    widget._on_button_pressed()
    assert widget.structure is None


@pytest.mark.usefixtures("aiida_profile_clean")
def test_smiles_canonicalization():
    """Test the SMILES canonicalization via RdKit."""
    canonicalize = awb.SmilesWidget.canonicalize_smiles

    # Should not change canonical smiles
    assert canonicalize("C") == "C"

    # Should canonicalize this
    canonical = canonicalize("O=CC=C")
    assert canonical == "C=CC=O"

    # Should be idempotent
    assert canonical == canonicalize(canonical)

    # Should raise for invalid smiles
    with pytest.raises(ValueError):
        canonicalize("invalid")
    # There is another failure mode when RDkit mol object is generated
    # but the canonicalization fails. I do not know how to trigger it though.


@pytest.mark.usefixtures("aiida_profile_clean")
def test_tough_smiles():
    widget = awb.SmilesWidget()
    assert widget.structure is None
    # Regression test for https://github.com/aiidalab/aiidalab-widgets-base/issues/505
    # Throwing in this non-canonical string should not raise
    widget.smiles.value = "C=CC1=C(C2=CC=C(C3=CC=CC=C3)C=C2)C=C(C=C)C(C4=CC=C(C(C=C5)=CC=C5C(C=C6C=C)=C(C=C)C=C6C7=CC=C(C(C=C8)=CC=C8C(C=C9C=C)=C(C=C)C=C9C%10=CC=CC=C%10)C=C7)C=C4)=C1"
    widget._on_button_pressed()
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "C72H54"

    # Regression test for https://github.com/aiidalab/aiidalab-widgets-base/issues/510
    widget.smiles.value = "CC1=C(C)C(C2=C3C=CC4=C(C5=C(C)C(C)=C(C6=C(C=CC=C7)C7=CC8=C6C=CC=C8)C(C)=C5C)C9=CC=C%10N9[Fe]%11(N%12C(C=CC%12=C(C%13=C(C)C(C)=C(C%14=C(C=CC=C%15)C%15=CC%16=C%14C=CC=C%16)C(C)=C%13C)C%17=CC=C2N%17%11)=C%10C%18=C(C)C(C)=C(C%19=C(C=CC=C%20)C%20=CC%21=C%19C=CC=C%21)C(C)=C%18C)N43)=C(C)C(C)=C1C%22=C(C=CC=C%23)C%23=CC%24=C%22C=CC=C%24"
    widget._on_button_pressed()
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "C116H92FeN4"


@pytest.mark.usefixtures("aiida_profile_clean")
def test_basic_cell_editor_widget(structure_data_object):
    """Test the `BasicCellEditor`."""
    widget = awb.BasicCellEditor()
    assert widget.structure is None

    # Set the structure.
    widget.structure = structure_data_object.get_ase()
    assert widget.structure.get_chemical_formula() == "Si2"

    # Convert to conventional cell.
    widget._to_conventional_cell()
    assert widget.structure.get_chemical_formula() == "Si8"

    # Convert to primitive cell.
    widget._to_primitive_cell()
    assert widget.structure.get_chemical_formula() == "Si2"

    # Change the cell parameters.
    widget.cell_parameters.children[2].children[1].value = 10.0
    widget._apply_cell_parameters()
    assert widget.structure.cell.cellpar()[2] == 10.0

    # make supercell using cell transformation
    widget.cell_transformation.children[0].children[0].value = 2
    widget._apply_cell_transformation()
    assert widget.structure.get_chemical_formula() == "Si4"
    # reset the cell transformation matrix
    widget._reset_cell_transformation_matrix()
    assert widget.cell_transformation.children[0].children[0].value == 1


@pytest.mark.usefixtures("aiida_profile_clean")
def test_basic_structure_editor(structure_data_object):
    """Test the `BasicStructureEditor`."""
    widget = awb.BasicStructureEditor()
    assert widget.structure is None

    # Set the structure.
    widget.structure = structure_data_object.get_ase()

    # Set first vector point vector to the first atom.
    widget.selection = [0]
    widget.def_axis_p1()
    assert widget.axis_p1.value == "0.0 0.0 0.0"

    # Set action point to the first atom.
    widget.def_point()
    assert widget.point.value == "0.0 0.0 0.0"

    # Set second vector point vector to the second atom.
    widget.selection = [1]
    widget.def_axis_p2()
    assert widget.axis_p2.value == "1.92 1.11 0.79"

    # Move an atom.
    original_position = widget.structure[
        1
    ].position  # original position of the second atom
    widget.displacement.value = "1.5"  # displacement in Angstrom
    assert np.allclose(
        widget.action_vector, np.array([0.8164966, 0.47140451, 0.33333329]), atol=1e-2
    )
    widget.translate_dr()
    assert np.linalg.norm(widget.structure[1].position - original_position) - 1.5 < 1e-6

    # Move an atom to its original position.
    widget.selection = [1]
    widget.dxyz.value = "{:.8f} {:.8f} {:.8f}".format(*original_position)
    widget.translate_to_xyz()
    assert np.allclose(widget.structure[1].position, original_position)

    # Move an atom by a vector (1, 1, 1).
    widget.selection = [1]
    widget.dxyz.value = "1.0 1.0 1.0"
    widget.translate_dxdydz()
    assert np.allclose(
        widget.structure[1].position, original_position + np.array([1.0, 1.0, 1.0])
    )

    # Copy an atom.
    widget.selection = [1]
    widget.copy_sel()
    assert len(widget.structure) == 3

    # Delete an atom.
    widget.selection = [2]
    widget.remove()
    assert len(widget.structure) == 2

    # Add an atom.
    widget.selection = [1]
    widget.element.value = "C"
    widget.add()
    assert len(widget.structure) == 3
    assert widget.structure[2].symbol == "C"

    # Modify an atom.
    widget.selection = [1]
    widget.element.value = "O"
    widget.mod_element()
    assert widget.structure.get_chemical_formula() == "COSi"

    # Add a ligand.
    widget.ligand.label = "Methyl -CH3"
    widget.selection = [2]
    widget.add()
    assert len(widget.structure) == 7
    assert widget.structure.get_chemical_formula() == "C2H3OSi"



"""Provide more user friendly error messages and automated reporting.

Authors:

    * Carl Simon Adorf <simon.adorf@epfl.ch>
"""

from __future__ import annotations

import base64
import json
import platform
import re
import subprocess
import sys
import textwrap
import zlib
from urllib import parse

import ipywidgets as ipw
from ansi2html import Ansi2HTMLConverter


def find_installed_packages(python_bin: str | None = None) -> dict[str, str]:
    """Return all currently installed packages."""
    if python_bin is None:
        python_bin = sys.executable
    output = subprocess.run(
        [python_bin, "-m", "pip", "list", "--format=json"],
        encoding="utf-8",
        capture_output=True,
        check=True,
    ).stdout

    return {package["name"]: package["version"] for package in json.loads(output)}


def get_environment_fingerprint(encoding="utf-8"):
    data = {
        "version": 1,
        "platform": {
            "architecture": platform.architecture(),
            "python_version": platform.python_version(),
            "version": platform.version(),
        },
        "packages": find_installed_packages(),
    }
    json_data = json.dumps(data, separators=(",", ":"))
    return base64.urlsafe_b64encode(zlib.compress(json_data.encode(encoding), level=9))


def parse_environment_fingerprint(fingerprint, encoding="utf-8"):
    """decode the environment fingerprint and return the data as a dictionary."""
    data = json.loads(
        zlib.decompress(base64.urlsafe_b64decode(fingerprint)).decode(encoding)
    )
    return data


ERROR_MESSAGE = """<div class="alert alert-danger">
<p><strong>
    <i class="fa fa-bug" aria-hidden="true"></i> Oh no... the application crashed due to an unexpected error.
</strong></p>
<a href="{issue_url}" target="_blank" class="btn btn-primary">
    <i class="fa fa-share" aria-hidden="true"></i> Create bug report
</a>
<button
    onclick="Jupyter.notebook.clear_all_output(); Jupyter.notebook.restart_run_all({{confirm: false}})"
    type="button"
    class="btn btn-success">
        <i class="fa fa-refresh" aria-hidden="true"></i> Restart app
</button>
<div style="padding-top: 1em">
    <details style="border: 1px solid #aaa; border-radius: 4px; padding: .5em .5em 0; ">
        <summary style="font-weight: bold; margin: -.5em -.5em 0; padding: .5em">
            <i class="fa fa-code" aria-hidden="true"></i> View the full traceback
        </summary>
      <pre style="color: #333; background: #f8f8f8;"><code>{traceback}</code></pre>
    </details>
</div></div>"""


BUG_REPORT_TITLE = """Bug report: Application crashed with {exception_type}"""

BUG_REPORT_BODY = """## Automated report

_This issue was created with the app's automated bug reporting feature.
Attached to this issue is the full traceback as well as an environment
fingerprint that contains information about the operating system as well as all
installed libraries._

## Additional comments (optional):

_Example: I submitted a band structure calculation for Silica._

## Attachments

<details>
<summary>Traceback</summary>

```python-traceback
{traceback}
```
</details>

<details>
<summary>Environment fingerprint</summary>
<pre>{environment_fingerprint}</pre>
</details>

**By submitting this issue I confirm that I am aware that this information can
potentially be used to determine what kind of calculation was performed at the
time of error.**
"""


def _strip_ansi_codes(msg):
    """Remove any ANSI codes (e.g. color codes)."""
    return re.sub(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])", "", msg)


def _convert_ansi_codes_to_html(msg):
    """Convert any ANSI codes (e.g. color codes) into HTML."""
    converter = Ansi2HTMLConverter()
    return converter.produce_headers().strip() + converter.convert(msg, full=False)


def _format_truncated_traceback(traceback, max_num_chars=2000):
    """Truncate the traceback to the given character length."""
    n = 0
    for _i, line in enumerate(reversed(traceback)):
        n += len(_strip_ansi_codes(line)) + 2  # add 2 for newline control characters
        if n > max_num_chars:
            break
    return _strip_ansi_codes("\n".join(traceback[-_i:]))


_ORIGINAL_EXCEPTION_HANDLER = None


def install_create_github_issue_exception_handler(output, url, labels=None):
    """Install a GitHub bug report exception handler.

    After installing this handler, kernel exception will show a generic error
    message to the user, with the option to file an automatic bug report at the
    given URL.

    This is an example of how to use this function:

    Example:
    --------
    .. highlight:: python
    .. code-block:: python

        output = ipw.Output()
        install_create_github_issue_exception_handler(
            output,
            url='https://github.com/aiidalab/aiidalab-qe/issues/new',
            labels=('bug', 'automated-report'))

        with output:
            display(welcome_message, app_with_work_chain_selector, footer)

    """
    global _ORIGINAL_EXCEPTION_HANDLER  # noqa

    if labels is None:
        labels = []

    ipython = get_ipython()  # noqa
    _ORIGINAL_EXCEPTION_HANDLER = _ORIGINAL_EXCEPTION_HANDLER or ipython._showtraceback

    def create_github_issue_exception_handler(exception_type, exception, traceback):
        try:
            output.clear_output()

            bug_report_query = {
                "title": BUG_REPORT_TITLE.format(
                    exception_type=str(exception_type.__name__)
                ),
                "body": BUG_REPORT_BODY.format(
                    # Truncate the traceback to a maximum of 2000 characters
                    # and strip all ansi control characters:
                    traceback=_format_truncated_traceback(traceback, 2000),
                    # Determine and format the environment fingerprint to be
                    # included with the bug report:
                    environment_fingerprint="\n".join(
                        textwrap.wrap(
                            get_environment_fingerprint().decode("utf-8"), 100
                        )
                    ),
                ),
                "labels": ",".join(labels),
            }
            issue_url = parse.urlunsplit(
                parse.urlsplit(url)._replace(query=parse.urlencode(bug_report_query))
            )

            with output:
                msg = ipw.HTML(
                    ERROR_MESSAGE.format(
                        issue_url=issue_url,
                        traceback=_convert_ansi_codes_to_html("\n".join(traceback)),
                        len_url=len(issue_url),
                    )
                )
                display(msg)  # noqa
        except Exception as error:
            print(f"Error while generating bug report: {error}", file=sys.stderr)
            _ORIGINAL_EXCEPTION_HANDLER(exception_type, exception, traceback)

    def restore_original_exception_handler():
        ipython._showtraceback = _ORIGINAL_EXCEPTION_HANDLER

    ipython._showtraceback = create_github_issue_exception_handler

    return restore_original_exception_handler



from pathlib import Path

import ase
import pytest
import traitlets as tl
from aiida import orm

from aiidalab_widgets_base import viewers


@pytest.mark.usefixtures("aiida_profile_clean")
def test_pbc_structure_data_viewer(structure_data_object):
    """Test the periodicity of the structure viewer widget."""
    # Prepare a structure with periodicity xy
    ase_input = ase.Atoms(
        symbols="Li2",
        positions=[(0.0, 0.0, 0.0), (1.5, 1.5, 1.5)],
        pbc=[True, True, False],
        cell=[3.5, 3.5, 3.5],
    )
    viewer = viewers.StructureDataViewer()
    viewer.structure = ase_input
    assert viewer.periodicity.value == "Periodicity: xy"
    assert viewer.cell_volume.value == "Cell area: 12.2500 (Å²)"


@pytest.mark.usefixtures("aiida_profile_clean")
def test_several_data_viewers(bands_data_object, generate_calc_job_node):
    v = viewers.viewer(orm.Int(1))

    # No viewer for Int, so it should return the input
    assert isinstance(v, orm.Int)

    # DictViewer
    v = viewers.viewer(orm.Dict(dict={"a": 1}))
    assert isinstance(v, viewers.DictViewer)

    # BandsDataViewer
    v = viewers.viewer(bands_data_object)
    assert isinstance(v, viewers.BandsDataViewer)

    # ProcessNodeViewer
    process = generate_calc_job_node(
        inputs={
            "parameters": orm.Int(1),
            "nested": {
                "inner": orm.Int(2),
            },
        }
    )
    v = viewers.viewer(process)
    assert isinstance(v, viewers.ProcessNodeViewerWidget)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_folder_data_viewer(folder_data_object):
    v = viewers.viewer(folder_data_object)
    assert isinstance(v, viewers.FolderDataViewer)

    v.files.value = "test1.txt"
    assert v.text.value == "content of test1.txt"

    v.files.value = "test2.txt"
    assert v.text.value == "content of test2.txt"
    v.download_btn.click()
    # NOTE: We're testing the download() method directly as well,
    # since triggering it via self.download_btn.click() callback
    # seems to swallow all exceptions.
    v.download()

    v.files.value = "test.bin"
    assert v.text.value == "[Binary file, preview not available]"
    v.download()


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_data_viewer_storage(monkeypatch, tmp_path, structure_data_object):
    v = viewers.viewer(structure_data_object)
    assert isinstance(v, viewers.StructureDataViewer)

    # Check the `_prepare_payload` function used for downloading.
    format_cases = {
        "Extended xyz": """MgpMYXR0aWNlPSIzLjg0NzM3IDAuMCAwLjAgMS45MjM2ODUgMy4zMzE5MiAwLjAgMS45MjM2ODUgMS4xMTA2NCAzLjE0MTM2NCIgUHJvcGVydGllcz1zcGVjaWVzOlM6MTpwb3M6UjozOm1hc3NlczpSOjE6X2FpaWRhbGFiX3ZpZXdlcl9yZXByZXNlbnRhdGlvbl9kZWZhdWx0Okk6MSBwYmM9IlQgVCBUIgpTaSAgICAgICAwLjAwMDAwMDAwICAgICAgIDAuMDAwMDAwMDAgICAgICAgMC4wMDAwMDAwMCAgICAgIDI4LjA4NTUwMDAwICAgICAgICAwClNpICAgICAgIDEuOTIzNjg1MDAgICAgICAgMS4xMTA2NDAwMCAgICAgICAwLjc4NTM0MTAwICAgICAgMjguMDg1NTAwMDAgICAgICAgIDAK""",
        "xsf": """Q1JZU1RBTApQUklNVkVDCiAzLjg0NzM3MDAwMDAwMDAwIDAuMDAwMDAwMDAwMDAwMDAgMC4wMDAwMDAwMDAwMDAwMAogMS45MjM2ODUwMDAwMDAwMCAzLjMzMTkyMDAwMDAwMDAwIDAuMDAwMDAwMDAwMDAwMDAKIDEuOTIzNjg1MDAwMDAwMDAgMS4xMTA2NDAwMDAwMDAwMCAzLjE0MTM2NDAwMDAwMDAwClBSSU1DT09SRAogMiAxCiAxNCAgICAgMC4wMDAwMDAwMDAwMDAwMCAgICAgMC4wMDAwMDAwMDAwMDAwMCAgICAgMC4wMDAwMDAwMDAwMDAwMAogMTQgICAgIDEuOTIzNjg1MDAwMDAwMDAgICAgIDEuMTEwNjQwMDAwMDAwMDAgICAgIDAuNzg1MzQxMDAwMDAwMDAK""",
        "cif": """ZGF0YV9pbWFnZTAKX2NoZW1pY2FsX2Zvcm11bGFfc3RydWN0dXJhbCAgICAgICBTaTIKX2NoZW1pY2FsX2Zvcm11bGFfc3VtICAgICAgICAgICAgICAiU2kyIgpfY2VsbF9sZW5ndGhfYSAgICAgICAzLjg0NzM3Cl9jZWxsX2xlbmd0aF9iICAgICAgIDMuODQ3MzY5ODYzMzc3NDQ4Cl9jZWxsX2xlbmd0aF9jICAgICAgIDMuODQ3MzY5NjE2OTM1ODM2Cl9jZWxsX2FuZ2xlX2FscGhhICAgIDU5Ljk5OTk5NzA5Nzk3MDEyCl9jZWxsX2FuZ2xlX2JldGEgICAgIDU5Ljk5OTk5NjcwNjQwOTMwNgpfY2VsbF9hbmdsZV9nYW1tYSAgICA1OS45OTk5OTg4MjUzMTc1OQoKX3NwYWNlX2dyb3VwX25hbWVfSC1NX2FsdCAgICAiUCAxIgpfc3BhY2VfZ3JvdXBfSVRfbnVtYmVyICAgICAgIDEKCmxvb3BfCiAgX3NwYWNlX2dyb3VwX3N5bW9wX29wZXJhdGlvbl94eXoKICAneCwgeSwgeicKCmxvb3BfCiAgX2F0b21fc2l0ZV90eXBlX3N5bWJvbAogIF9hdG9tX3NpdGVfbGFiZWwKICBfYXRvbV9zaXRlX3N5bW1ldHJ5X211bHRpcGxpY2l0eQogIF9hdG9tX3NpdGVfZnJhY3RfeAogIF9hdG9tX3NpdGVfZnJhY3RfeQogIF9hdG9tX3NpdGVfZnJhY3RfegogIF9hdG9tX3NpdGVfb2NjdXBhbmN5CiAgU2kgIFNpMSAgICAgICAxLjAgIDAuMCAgMC4wICAwLjAgIDEuMDAwMAogIFNpICBTaTIgICAgICAgMS4wICAwLjI1MDAwMDAwMDAwMDAwMDA2ICAwLjI1ICAwLjI1ICAxLjAwMDAK""",
    }

    # Compatibility with old ASE versions <3.23
    old_ase = format_cases.copy()
    old_ase["cif"] = (
        """ZGF0YV9pbWFnZTAKX2NoZW1pY2FsX2Zvcm11bGFfc3RydWN0dXJhbCAgICAgICBTaTIKX2NoZW1pY2FsX2Zvcm11bGFfc3VtICAgICAgICAgICAgICAiU2kyIgpfY2VsbF9sZW5ndGhfYSAgICAgICAzLjg0NzM3Cl9jZWxsX2xlbmd0aF9iICAgICAgIDMuODQ3MzcKX2NlbGxfbGVuZ3RoX2MgICAgICAgMy44NDczNwpfY2VsbF9hbmdsZV9hbHBoYSAgICA2MApfY2VsbF9hbmdsZV9iZXRhICAgICA2MApfY2VsbF9hbmdsZV9nYW1tYSAgICA2MAoKX3NwYWNlX2dyb3VwX25hbWVfSC1NX2FsdCAgICAiUCAxIgpfc3BhY2VfZ3JvdXBfSVRfbnVtYmVyICAgICAgIDEKCmxvb3BfCiAgX3NwYWNlX2dyb3VwX3N5bW9wX29wZXJhdGlvbl94eXoKICAneCwgeSwgeicKCmxvb3BfCiAgX2F0b21fc2l0ZV90eXBlX3N5bWJvbAogIF9hdG9tX3NpdGVfbGFiZWwKICBfYXRvbV9zaXRlX3N5bW1ldHJ5X211bHRpcGxpY2l0eQogIF9hdG9tX3NpdGVfZnJhY3RfeAogIF9hdG9tX3NpdGVfZnJhY3RfeQogIF9hdG9tX3NpdGVfZnJhY3RfegogIF9hdG9tX3NpdGVfb2NjdXBhbmN5CiAgU2kgIFNpMSAgICAgICAxLjAgIDAuMDAwMDAgIDAuMDAwMDAgIDAuMDAwMDAgIDEuMDAwMAogIFNpICBTaTIgICAgICAgMS4wICAwLjI1MDAwICAwLjI1MDAwICAwLjI1MDAwICAxLjAwMDAK"""
    )

    for fmt in format_cases:  # noqa: PLC0206
        v.file_format.label = fmt
        b64_payload = v._prepare_payload()
        try:
            assert b64_payload == format_cases[fmt], (
                f"{fmt} structure payload does not match"
            )
        except AssertionError:
            assert b64_payload == old_ase[fmt], (
                f"{fmt} structure payload does not match"
            )

    # Monkey patch the viewer to avoid the need for a running X server.
    # fmt: off
    v._viewer._camera_orientation = [
        16.619212980943573, 0, 0, 0,
        0, 16.619212980943573, 0, 0,
        0, 0, 16.619212980943573, 0,
        -1.6859999895095825, -1.6859999895095825, -0.6669999957084656, 1,
    ]
    # fmt: on
    # Avoid producing temporary files from povray in the repo
    monkeypatch.chdir(tmp_path)
    v._render_structure()

    # Make sure we don't polute current working dir with tempfiles
    assert not Path("__temp__.pov").exists()
    assert not Path("Si2.png").exists()


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_data_viewer_selection(structure_data_object):
    v = viewers.viewer(structure_data_object)

    # Direct selection.
    v._selected_atoms.value = "1..2"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [0, 1]
    assert "Distance" in v.selection_info.value

    # The x coordinate lower than 0.5.
    v._selected_atoms.value = "x<0.5"
    v.apply_displayed_selection()
    assert v.selection == [0]
    assert v.displayed_selection == [0]

    # The id of the second atom
    v._selected_atoms.value = "id > 1"
    v.apply_displayed_selection()
    assert v.selection == [1]

    # or of the two selections.
    v._selected_atoms.value = "x>=0.5 or x<0.5"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]

    # Display 2*2*2 supercell
    v.supercell = [2, 2, 2]
    assert len(v.structure) == 2
    assert len(v.displayed_structure) == 16

    # Test intersection of the selection with the supercell.
    v._selected_atoms.value = "z>0 and z<2.5"
    v.apply_displayed_selection()
    assert v.selection == [1]
    assert v.displayed_selection == [1, 5, 9, 13]

    v._selected_atoms.value = "x<=2.0 and z<3"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [4, 0, 1]
    assert "Angle" in v.selection_info.value

    # Convert to boron nitride.
    new_structure = v.structure.copy()
    new_structure.symbols = ["B", "N"]
    v.structure = None
    v.structure = new_structure

    # Use "name" and "not" operators.
    v._selected_atoms.value = "z<2 and name B"
    v.apply_displayed_selection()
    assert v.selection == [0]
    assert v.displayed_selection == [0, 4, 8, 12]

    v._selected_atoms.value = "z<2 and name not B"
    v.apply_displayed_selection()
    assert v.selection == [1]
    assert v.displayed_selection == [1, 5, 9, 13]

    v._selected_atoms.value = "z<2 and name not [B, O]"
    v.apply_displayed_selection()
    assert v.selection == [1]
    assert v.displayed_selection == [1, 5, 9, 13]

    # Use "id" operator.
    v._selected_atoms.value = "id == 1 or id == 8"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [0, 7]

    # Use the d_from operator.
    v._selected_atoms.value = "d_from[0,0,0] < 4"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [4, 8, 0, 1, 2]

    # Use the != operator.
    v._selected_atoms.value = "id != 5"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]

    # Use ^ and - operators.
    v._selected_atoms.value = "(x-4)^2 + (y-2)^2 < 4"
    v.apply_displayed_selection()
    assert v.selection == [1, 0]
    assert v.displayed_selection == [3, 9, 10]

    # Division and multiplication.
    v._selected_atoms.value = "x/2 < 1"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [4, 0, 1, 2]

    v._selected_atoms.value = "x*1.5 < y + z"
    v.apply_displayed_selection()
    assert v.selection == [0, 1]
    assert v.displayed_selection == [2, 3, 4, 6, 7]

    # Test wrong syntax.
    assert v.wrong_syntax.layout.visibility == "hidden"
    v._selected_atoms.value = "x--x"
    v.apply_displayed_selection()
    assert v.wrong_syntax.layout.visibility == "visible"


@pytest.mark.usefixtures("aiida_profile_clean")
def test_structure_data_viewer_representation(structure_data_object):
    v = viewers.viewer(structure_data_object)

    # By default, there should be one "default" representation.
    assert len(v._all_representations) == 1
    assert (
        v._all_representations[0].style_id == "_aiidalab_viewer_representation_default"
    )
    assert v._all_representations[0].selection.value == "1..2"

    # Display only one atom.
    v._all_representations[0].selection.value = "1"
    v._apply_representations()
    assert "2" in v.atoms_not_represented.value

    # Add a new representation.
    v._add_representation()
    assert "2" in v.atoms_not_represented.value
    v._all_representations[1].selection.value = "2"
    v._all_representations[0].type.value = "ball+stick"
    v._all_representations[1].type.value = "spacefill"
    v._apply_representations()
    assert v.atoms_not_represented.value == ""

    # Add an atom to the structure.
    new_structure = v.structure.copy()
    new_structure.append(ase.Atom("C", (0.5, 0.5, 0.5)))
    v.structure = None
    v.structure = new_structure

    # The new atom should appear in the default representation.
    assert v._all_representations[0].selection.value == "1 3"
    assert "3" not in v.atoms_not_represented.value

    # Delete the second representation.
    assert v._all_representations[0].delete_button.layout.visibility == "hidden"
    assert v._all_representations[1].delete_button.layout.visibility == "visible"
    v._all_representations[1].delete_button.click()
    assert len(v._all_representations) == 1
    assert "2" in v.atoms_not_represented.value

    # Try to provide different object type than the viewer accepts.
    with pytest.raises(tl.TraitError):
        v.structure = 2

    with pytest.raises(tl.TraitError):
        v.structure = orm.Int(1)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_compute_bonds_in_structure_data_viewer():
    # Check the function to compute bonds.
    water = ase.Atoms(
        symbols=["O", "H", "H"],
        positions=[
            (0.0, 0.0, 0.119262),
            (0.0, 0.763239, -0.477047),
            (0.0, -0.763239, -0.477047),
        ],
    )
    viewer = viewers.StructureDataViewer()
    bonds = viewer._compute_bonds(water)
    assert len(bonds) == 4


@pytest.mark.usefixtures("aiida_profile_clean")
def test_loading_viewer_using_process_type(generate_calc_job_node):
    """Test loading a viewer widget based on the process type of the process node."""
    from aiidalab_widgets_base import register_viewer_widget

    # Define and register a viewer widget for the calculation type identified by "aiida.calculations:abc".
    @register_viewer_widget("aiida.calculations:abc")
    class AbcViewer:
        def __init__(self, node=None):
            self.node = node

    # Generate a calc job node with the specific entry point "abc".
    process = generate_calc_job_node(entry_point_name="abc")
    # Load the viewer widget for the generated process node.
    viewer = viewers.viewer(process)
    # Verify that the loaded viewer is the correct type and is associated with the intended node.
    assert isinstance(viewer, AbcViewer), (
        "Viewer is not an instance of the expected viewer class."
    )
    assert viewer.node == process, "Viewer's node does not match the test process node."


def test_node_view_for_non_widget_viewer():
    """Test that a node with no registered viewer is displayed in an output widget"""
    import sys
    from io import StringIO

    # Intercepting stdout because `ipw.Output` does not
    # store outputs in non-interactive environments.
    captured = StringIO()
    sys.stdout = captured

    node_view = viewers.AiidaNodeViewWidget()
    node = orm.Int(1)
    node_view.node = node
    assert node_view.children[0] is node_view._output
    assert str(node) in sys.stdout.getvalue()


def test_node_view_caching():
    """Test that providing a given node a second time returns the cached viewer."""
    node_view = viewers.AiidaNodeViewWidget()
    node = orm.Int(1)
    node_view.node = node
    viewer = node_view.children[0]
    node_view.node = None
    node_view.node = node
    assert node_view.children[0] is viewer



from pathlib import Path

from aiidalab_widgets_base.loaders import LoadingWidget
from aiidalab_widgets_base.utils.loaders import load_css


def test_load_css():
    """Test `load_css` utility."""
    css_dir = Path("aiidalab_widgets_base/static/styles")
    load_css(css_path=css_dir)
    load_css(css_path=css_dir / "global.css")


def test_loading_widget():
    """Test `LoadingWidget`."""
    widget = LoadingWidget(message="Loading some widget")
    assert widget.message.value == "Loading some widget"
    assert widget.children[0].value == "Loading some widget"
    assert widget.children[1].value == "<i class='fa fa-spinner fa-spin fa-2x fa-fw'/>"
    assert "loading" in widget._dom_classes
